# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DataTransformer classes that handle the transformations generated on input
data by Feature Engineering, Models, Anomaly Detectors, etc.
The BMLDataTransformer will take care of transformations that use BigML
objects as transformation generators. Other libraries, like Pandas
(DFDataTransfomer) and scikit-learn (SKDataTransformer)
will need their own DataTransformer subclasses to define
their own transformations.

"""

import types

from datetime import datetime

from bigml.constants import INTERNAL, DATAFRAME, OUT_NEW_HEADERS
from bigml.util import get_formatted_data, format_data, get_data_format

try:
    from pandas import DataFrame, concat
    PANDAS_READY = True
except ImportError:
    PANDAS_READY = False


class DataTransformer():
    """Base class to handle transformations. It offers a transform method
    that can handle list of dictionaries or Pandas DataFrames a inputs and
    delegates to the `data_transform` method the actual transformations to
    be applied and should be implemented in the classes derived from it.
    """

    def __init__(self, generator, data_format, resource_id=None, name=None,
                 description=None):
        """Adds initial attributes:
         - generator: object, function or list of functions that will be
           doing the transformation
         - data_format: whether to accept a DataFrame or a list of dictionaries
           as inputs for the generator
         - resource_id: unique identifier for the data transformer object
         - name: name for the data transformer
         - description: description for the transformations in the data
                        transformer
        """
        self.generator = generator
        self.data_format = data_format
        self.resource_id = resource_id
        self.name = name
        self.description = description

    def _formatted_input(self, input_data_list):
        """Returns a copy of the input data list in the expected format """
        return get_formatted_data(input_data_list, self.data_format)

    def transform(self, input_data_list, out_format=None):
        """Returns a new input_data_list where the transformations defined
        in the generator have been applied. It handles format transformation
        if needed before applying the generator function.
        """
        data_format = get_data_format(input_data_list)
        inner_data_list = self._formatted_input(input_data_list)
        result = self.data_transform(inner_data_list)
        if self.data_format != data_format and out_format is None:
            return format_data(result, data_format)
        if self.data_format != out_format:
            return format_data(result, out_format)
        return result

    def data_transform(self, input_data_list):
        """Method to be re-implemented in each of the data transformers. Using
        identity by default."""
        raise NotImplementedError("This method needs to be implemented")


class BMLDataTransformer(DataTransformer):
    """Transformer wrapper for BigML resources."""
    def __init__(self, local_resource, outputs=None, **kwargs):
        """Receives a local resource (Dataset, SupervisedModel, Cluster...)
        and creates a `DataTransformer` from it to apply the corresponding
        transformations.
        - for Datasets, Flatline transformations (if any) are applied
        - for models, a batch prediction (scoring, topic distribution, etc.) is
          applied and added to the original input.

        Optional arguments are:
        :param outputs: dictionary of output fields and headers
        :type outputs: dict
        :param kwargs: dictionary of runtime settings for batch predictions
                       (e.g. missing_strategy, operating_point, etc.)
        :type kwargs: dict
        """
        try:
            generator = local_resource.transform
            self.add_input = False
        except AttributeError:
            if hasattr(local_resource, "batch_predict"):
                generator = lambda x : \
                    local_resource.batch_predict(x, outputs=outputs, **kwargs)
                self.add_input = True
            else:
                raise ValueError("The local resource needs to provide "
                                 "a transform, or batch_predict "
                                 "method to generate transformations.")
        super().__init__(generator,
                         INTERNAL,
                         local_resource.resource_id,
                         local_resource.name,
                         local_resource.description)
        self.local_resource = local_resource
        self.dump = local_resource.dump

    def data_transform(self, input_data_list):
        """Returns a list of dictionaries with the generated transformations.
        The input list is expected to be a list of dictionaries"""
        return self.generator(input_data_list)

    def merge_input_data(self, input_data_list, output_data_list,
                         out_format=None):
        """Adding input data to the output """
        data_format = get_data_format(input_data_list)
        input_data_list = self._formatted_input(input_data_list)
        output_data_list = self._formatted_input(output_data_list)
        for index, input_data in enumerate(input_data_list):
            for key, value in input_data.items():
                if key not in output_data_list[index]:
                    output_data_list[index].update({key: value})
        if self.data_format != out_format:
            return format_data(output_data_list, data_format)
        return output_data_list


class DFDataTransformer(DataTransformer):
    """DataTransformer wrapper for DataFrames """
    def __init__(self, generator, resource_id=None, name=None,
                 description=None):
        """Receives the function or list of functions to be applied on
        the input DataFrame
        Optional parameters are:
        :param resource_id: unique ID for the DataTransformer
        :type resource_id: str
        :param name: DataTransformer name
        :type name: str
        :param description: Description for the transformations.
        :type description: str
        """
        if not isinstance(generator, list):
            generator = [generator]
        for index, item in enumerate(generator):
            if not isinstance(item, tuple) and isinstance(
                    item, types.FunctionType):
                generator[index] = (item, [], {})
            elif isinstance(item, tuple) and isinstance(
                    item[0], types.FunctionType):
                try:
                    args = item[1]
                    if not isinstance(args, list):
                        raise ValueError("The syntax of the first argument is "
                                         " function or (function, list, dict)")
                except IndexError:
                    args = []
                try:
                    kwargs = item[2]
                    if not isinstance(kwargs, dict):
                        raise ValueError("The syntax of the first argument is "
                                         " function or (function, list, dict)")
                except IndexError:
                    kwargs = {}

                generator[index] = (item[0], args, kwargs)
            else:
                raise ValueError("Only functions or tuples of functions are "
                                 "allowed as first argument.")

        super().__init__(generator,
                         DATAFRAME,
                         resource_id or "dftrans_%s" %
                             str(datetime.now()).replace(" ", "_"),
                         name,
                         description)

    def data_transform(self, input_data_list):
        """Calling the corresponding method in the generator.
        The input_data_list is expected to be a Dataframe.

        """
        result = input_data_list.copy()
        for function, args, kwargs in self.generator:
            result = result.pipe(function, *args, **kwargs)
        return result


class SKDataTransformer(DataTransformer):
    """DataTransformer wrapper for scikit learn pipelines or transformations """
    def __init__(self, generator, resource_id=None, name=None,
                 description=None, output=None):
        """Receives the pipeline or transformation to be applied on
        the input DataFrame
        Optional parameters are:
        :param resource_id: unique ID for the DataTransformer
        :type resource_id: str
        :param name: DataTransformer name
        :type name: str
        :param description: Description for the transformations.
        :type description: str
        :param output: Dictionary containing the headers to be used for the
                       new fields generated in the transformation.
        :type output: dict
        """

        try:
            generator_fn = generator.transform
            self.add_input = False
        except AttributeError:
            try:
                generator_fn = generator.predict
                self.add_input = True
            except AttributeError:
                try:
                    generator_fn = generator.score
                    self.add_input = True
                except AttributeError:
                    raise ValueError("Failed to find a .transform, .predict "
                                     "or .score method in the first argument "
                                     "object.")

        super().__init__(generator_fn,
                         DATAFRAME,
                         resource_id or "sktrans_%s" %
                             str(datetime.now()).replace(" ", "_"),
                         name,
                         description)
        self.output = output or {}
        try:
            self.output_headers = generator.get_feature_names_out()
        except AttributeError:
            self.output_headers = self.output.get(OUT_NEW_HEADERS)

    def data_transform(self, input_data_list):
        """Calling the corresponding method in the generator.
        The input_data_list is expected to be a Dataframe.

        """
        result = self.generator(input_data_list)
        try:
            result = result.toarray()
        except AttributeError:
            pass
        df_kwargs = {"index": input_data_list.index}
        if self.output_headers is not None:
            df_kwargs.update({"columns": self.output_headers})
        result = DataFrame(result, **df_kwargs)
        if not self.add_input:
            return result
        return concat([input_data_list, result], axis=1)

    @staticmethod
    def merge_input_data(input_data_list, output_data_list):
        """Adding input data to the output """
        return concat([input_data_list, output_data_list], axis=1)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Comparing remote and local predictions

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_model_steps as model_create
from . import create_time_series_steps as time_series_create
from . import create_forecast_steps as forecast_create
from . import compare_forecasts_steps as forecast_compare
from . import create_pca_steps as pca_create
from . import create_projection_steps as projection_create
from . import compare_predictions_steps as compare_predictions


class TestComparePrediction:
    """Test local and remote predictions"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario2(self):
        """
        Scenario: Successfully comparing forecasts from time series with "A" seasonality
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a time series with "<model_conf>"
            And I wait until the time series is ready less than <model_wait> secs
            And I create a local time series
            When I create a forecast for "<input_data>"
            Then the forecast is "<forecast>"
            And I create a local forecast for "<input_data>"
            Then the local forecast is "<forecast>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "forecast", "model_conf"]
        examples = [

            ['data/grades.csv', '30', '30', '120',
             '{"000005": {"horizon": 5}}',
             '{"000005": [{"point_forecast": [73.96192, 74.04106, 74.12029, '
             '74.1996, 74.27899], "model": "M,M,N"}]}',
             '{"objective_fields": ["000001", "000005"], "period": 12}'],
            ['data/grades.csv', '30', '30', '120',
             '{"000005": {"horizon": 5, "ets_models": {"names": ["M,N,A"], '
             '"criterion": "aic", "limit": 3}}}',
             '{"000005": [{"point_forecast":  [67.43222, 68.24468, '
             '64.14437, 67.5662, 67.79028], "model": "M,N,A"}]}',
             '{"objective_fields": ["000001", "000005"], "period": 12}'],
            ['data/grades.csv', '30', '30', '120',
             '{"000005": {"horizon": 5, "ets_models": {"names": ["A,A,A"], '
             '"criterion": "aic", "limit": 3}}}',
             '{"000005": [{"point_forecast": [74.73553, 71.6163, 71.90264, '
             '76.4249, 75.06982], "model": "A,A,A"}]}',
             '{"objective_fields": ["000001", "000005"], "period": 12}']]
        show_doc(self.test_scenario2)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            time_series_create.i_create_a_time_series_with_params(
                self, example["model_conf"])
            time_series_create.the_time_series_is_finished_in_less_than(
                self, example["model_wait"])
            time_series_create.create_local_time_series(self)
            forecast_create.i_create_a_forecast(self, example["input_data"])
            forecast_create.the_forecast_is(self, example["forecast"])
            forecast_compare.i_create_a_local_forecast(
                self, example["input_data"])
            forecast_compare.the_local_forecast_is(self, example["forecast"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

from bigml.api import HTTP_OK
from bigml.api import HTTP_CREATED
from bigml.api import HTTP_ACCEPTED
from bigml.api import FINISHED
from bigml.api import FAULTY
from bigml.api import get_status
from bigml.api import BigML
from bigml.model import Model
from bigml.logistic import LogisticRegression
from bigml.linear import LinearRegression
from bigml.deepnet import Deepnet
from bigml.fusion import Fusion
from bigml.ensemble import Ensemble


from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_, ok_


NO_MISSING_SPLITS = {'missing_splits': False}


def i_create_a_model(step, shared=None):
    """Step: I create a model"""
    if shared is None or world.shared.get("model", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_model(dataset, args=NO_MISSING_SPLITS)
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.model = resource['object']
        world.models.append(resource['resource'])


def i_export_model(step, pmml, filename):
    """Step: I export the <pmml> model to file <filename>"""
    world.api.export(world.model["resource"], res_filename(filename), pmml)


def i_export_tags_model(step, filename, tag):
    """Step: I export the last model"""
    world.api.export_last(tag,
                          filename=res_filename(filename))


def i_create_a_balanced_model(step):
    """Step: I create a balanced model"""
    dataset = world.dataset.get('resource')
    args = {}
    args.update(NO_MISSING_SPLITS)
    args.update({"balance_objective": True})
    resource = world.api.create_model(dataset, args=args)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.model = resource['object']
    world.models.append(resource['resource'])


def i_create_a_model_from_dataset_list(step):
    """Step: I create a model from a dataset list"""
    resource = world.api.create_model(step.bigml["dataset_ids"],
                                      args=NO_MISSING_SPLITS)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.model = resource['object']
    world.models.append(resource['resource'])


def wait_until_model_status_code_is(step, code1, code2, secs):
    """Step: I wait until the model status code is either
    <code1> or <code2> less than <secs>
    """
    wait_until_status_code_is(code1, code2, secs, world.model)


def the_model_is_finished_in_less_than(step, secs, shared=None):
    """Step: I wait until the model is ready less than <secs>"""
    if shared is None or world.shared.get("model", {}).get(shared) is None:
        wait_until_model_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "model" not in world.shared:
                world.shared["model"] = {}
            world.shared["model"][shared] = world.model
        print("New %s" % world.model["resource"])
    else:
        world.model = world.shared["model"][shared]
        print("Reusing %s" % world.model["resource"])


def i_create_a_model_with(step, data="{}"):
    """Step: I create a model with <data>"""
    args = json.loads(data)
    if not 'missing_splits' in args:
        args.update(NO_MISSING_SPLITS)
    resource = world.api.create_model(world.dataset.get('resource'),
                                      args=args)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.model = resource['object']
    world.models.append(resource['resource'])


def i_create_a_model_with_missing_splits(step):
    """Step: I create a model with missing splits"""
    i_create_a_model_with(step, data='{"missing_splits": true}')


def i_create_a_weighted_model_with_missing_splits(step):
    """Step: I create a model with missing splits"""
    i_create_a_model_with(step, data='{"missing_splits": true, "balance_objective": true}')


def make_the_model_public(step):
    """Step: I make the model public"""
    resource = world.api.update_model(world.model['resource'],
                                      {'private': False, 'white_box': True})
    world.status = resource['code']
    if world.status != HTTP_ACCEPTED:
        print("unexpected status: %s" % world.status)
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.model = resource['object']


def model_from_public_url(step):
    """Step: I check the model status using the model''s public url"""
    world.model = world.api.get_model("public/%s" % world.model['resource'])
    eq_(get_status(world.model)['code'], FINISHED)


def make_the_model_shared(step):
    """Step: I make the model shared"""
    resource = world.api.update_model(world.model['resource'],
                                      {'shared': True})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.model = resource['object']


def get_sharing_info(step):
    """Step: I get the model sharing info"""
    world.shared_hash = world.model['shared_hash']
    world.sharing_key = world.model['sharing_key']


def model_from_shared_url(step):
    """Step: I check the model status using the model's shared url"""
    world.model = world.api.get_model("shared/model/%s" % world.shared_hash)
    eq_(get_status(world.model)['code'], FINISHED)


def model_from_shared_key(step):
    """Step: I check the model status using the model's shared key"""
    username = os.environ.get("BIGML_USERNAME")
    world.model = world.api.get_model(world.model['resource'],
        shared_username=username, shared_api_key=world.sharing_key)
    eq_(get_status(world.model)['code'], FINISHED)


def field_name_to_new_name(step, field_id, new_name):
    """Step: <field_id> field's name is changed to <new_name>"""
    eq_(step.bigml["local_model"].fields[field_id]['name'], new_name)


def i_create_a_model_from_cluster(step, centroid_id):
    """Step: I create a model associated to centroid <centroid_id>"""
    resource = world.api.create_model(
        world.cluster['resource'],
        args={'centroid': centroid_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.model = resource['object']
    world.models.append(resource['resource'])


def is_associated_to_centroid_id(step, centroid_id):
    """Step: the model is associated to the centroid <centroid_id> of the
    cluster
    """
    cluster = world.api.get_cluster(world.cluster['resource'])
    world.status = cluster['code']
    eq_(world.status, HTTP_OK)
    eq_("model/%s" % (cluster['object']['cluster_models'][centroid_id]),
        world.model['resource'])


def i_create_a_logistic_model(step, shared=None):
    """Step: I create a logistic regression model"""
    if shared is None or world.shared.get("logistic", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_logistic_regression(dataset)
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.logistic_regression = resource['object']
        world.logistic_regressions.append(resource['resource'])


def i_create_a_logistic_model_with_objective_and_parms(step, objective=None,
                                                       parms=None):
    """Step: I create a logistic regression model with objective <objective>
    and parms <parms>
    """
    dataset = world.dataset.get('resource')
    if parms is None:
        parms = {}
    else:
        parms = json.loads(parms)
    if objective is not None:
        parms.update({"objective_field": objective})
    resource = world.api.create_logistic_regression( \
        dataset, parms)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.logistic_regression = resource['object']
    world.logistic_regressions.append(resource['resource'])

def wait_until_logistic_model_status_code_is(step, code1, code2, secs):
    """Step: I wait until the logistic regression model status code is either
    <code1> or <code2> less than <secs>
    """
    world.logistic_regression = wait_until_status_code_is(
        code1, code2, secs, world.logistic_regression)


def the_logistic_model_is_finished_in_less_than(step, secs, shared=None):
    """Step: I wait until the logistic regression model is ready less than
        <secs>
    """
    if shared is None or world.shared.get("logistic", {}).get(shared) is None:
        wait_until_logistic_model_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "logistic" not in world.shared:
                world.shared["logistic"] = {}
            world.shared["logistic"][shared] = world.logistic_regression
    else:
        world.logistic_regression = world.shared["logistic"][shared]
        print("Reusing %s" % world.logistic_regression["resource"])


def i_create_a_deepnet(step, shared=None):
    """Step: I create a deepnet model"""
    if shared is None or world.shared.get("deepnet", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_deepnet(dataset)
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.deepnet = resource['object']
        world.deepnets.append(resource['resource'])


def i_create_a_quick_deepnet(step):
    """Step: I create a quick deepnet"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_deepnet(dataset, {"max_training_time": 100})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.deepnet = resource['object']
    world.deepnets.append(resource['resource'])


def i_create_a_no_suggest_deepnet(step, shared=None):
    """Step: I create a non-suggested deepnet model"""
    if shared is None or \
            world.shared.get("deepnet", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_deepnet(dataset, {"suggest_structure": False,
                                                      "max_iterations": 100,
                                                      "deepnet_seed": "bigml"})
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.deepnet = resource['object']
        world.deepnets.append(resource['resource'])


def i_create_a_deepnet_with_objective_and_params(step, objective=None, parms=None):
    """Step: I create a deepnet model with objective <objective> and parms
    <parms>
    """
    dataset = world.dataset.get('resource')
    if parms is None:
        parms = {}
    else:
        parms = json.loads(parms)
    if objective is not None:
        parms.update({"objective_field": objective})
    resource = world.api.create_deepnet(dataset, parms)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.deepnet = resource['object']
    world.deepnets.append(resource['resource'])


def wait_until_deepnet_model_status_code_is(step, code1, code2, secs):
    """Step: I wait until the deepnet model status code is either <code1>
    or <code2> less than <secs>
    """
    world.deepnet = wait_until_status_code_is(code1, code2, secs, world.deepnet)


def the_deepnet_is_finished_in_less_than(step, secs, shared=None):
    """Step: wait until the deepnet model is ready less than <secs>"""
    if shared is None or world.shared.get("deepnet", {}).get(shared) is None:
        wait_until_deepnet_model_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "deepnete" not in world.shared:
                world.shared["deepnet"] = {}
            world.shared["deepnet"][shared] = world.deepnet
    else:
        world.deepnet = world.shared["deepnet"][shared]
        print("Reusing %s" % world.deepnet["resource"])


def i_check_model_stored(step, filename, pmml):
    """Step: I check the model is stored in <filename> file in <pmml>"""
    with open(res_filename(filename)) as file_handler:
        content = file_handler.read()
        model_id = world.model["resource"][ \
            (world.model["resource"].index("/") + 1):]
        ok_(content.index(model_id) > -1)


def i_read_model_file(step, filename):
    """Step: I read model from file <filename>"""
    with open(res_filename(filename)) as file_handler:
        content = file_handler.read()
        world.model = json.loads(content)


def i_create_an_optiml(step):
    """Step: I create an optiml"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_optiml(dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.optiml = resource['object']
    world.optimls.append(resource['resource'])


def i_create_an_optiml_with_objective_and_params(step, objective=None, parms=None):
    """Step: I create an optiml model with objective <objective> and parms
    <parms>
    """
    dataset = world.dataset.get('resource')
    if parms is None:
        parms = {}
    else:
        parms = json.loads(parms)
    if objective is not None:
        parms.update({"objective_field": objective})
    resource = world.api.create_optiml(dataset, parms)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.optiml = resource['object']
    world.optimls.append(resource['resource'])


def wait_until_optiml_status_code_is(step, code1, code2, secs):
    """Step: I wait until the optiml status code is either <code1> or
    <code2> less than <secs>
    """
    world.optiml = wait_until_status_code_is(code1, code2, secs, world.optiml)


def the_optiml_is_finished_in_less_than(step, secs):
    """Step: I wait until the optiml is ready less than <secs>"""
    wait_until_optiml_status_code_is(step, FINISHED, FAULTY, secs)


def i_update_optiml_name(step, name):
    """Step: I update the optiml name to <name>"""
    resource = world.api.update_optiml(world.optiml['resource'],
                                      {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.optiml = resource['object']


def i_check_optiml_name(step, name):
    """Step: the optiml name is <name>"""
    optiml_name = world.optiml['name']
    eq_(name, optiml_name)


def i_create_a_fusion(step):
    """Step: I create a fusion"""
    resource = world.api.create_fusion(world.list_of_models,
                                       {"project": world.project_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.fusion = resource['object']
    world.fusions.append(resource['resource'])


def i_create_a_fusion_with_weights(step, weights=None):
    """Step: I create a fusion with weights"""
    if weights is None:
        weights = list(range(1, len(world.list_of_models)))
    else:
        weights = json.loads(weights)
    models = []
    try:
        for index, model in enumerate(world.list_of_models):
            models.append({"id": model["resource"], "weight": weights[index]})
    except IndexError:
        pass
    resource = world.api.create_fusion(models,
                                       {"project": world.project_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.fusion = resource['object']
    world.fusions.append(resource['resource'])


def i_create_a_fusion_with_objective_and_params(step, objective, parms=None):
    """Step: I create a fusion with objective <objective> and parms <parms>"""
    models = world.list_models
    if parms is None:
        parms = {}
    else:
        parms = json.loads(parms)
    parms.update({"objective_field": objective, "project": world.project_id})
    resource = world.api.create_fusion(models, parms)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.fusion = resource['object']
    world.fusions.append(resource['resource'])


def wait_until_fusion_status_code_is(step, code1, code2, secs):
    """Step: I wait until the fusion status code is either <code1> or
    <code2> less than <secs>
    """
    world.fusion = wait_until_status_code_is(code1, code2, secs, world.fusion)


def the_fusion_is_finished_in_less_than(step, secs):
    """Step: I wait until the fusion is ready less than <secs>"""
    wait_until_fusion_status_code_is(step, FINISHED, FAULTY, secs)


def i_update_fusion_name(step, name):
    """Step: I update the fusion name to <name>"""
    resource = world.api.update_fusion(world.fusion['resource'],
                                       {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.fusion = resource['object']


def i_check_fusion_name(step, name):
    """Step: the fusion name is <name>"""
    fusion_name = world.fusion['name']
    eq_(name, fusion_name)


def i_create_local_model_from_file(step, export_file):
    """Step: I create a local model from file <export_file>"""
    step.bigml["local_model"] = Model( \
        res_filename(export_file),
        api=BigML("wrong-user", "wrong-api-key"))


def check_model_id_local_id(step):
    """Step: the model ID and the local model ID match"""
    eq_(step.bigml["local_model"].resource_id, world.model["resource"])


def i_export_ensemble(step, filename):
    """Step: I export the ensemble"""
    world.api.export(world.ensemble.get('resource'),
                     filename=res_filename(filename))


def i_create_local_ensemble_from_file(step, export_file):
    """Step: I create a local ensemble from file <export_file>"""
    step.bigml["local_ensemble"] = Ensemble( \
        res_filename(export_file),
        api=BigML("wrong-user", "wrong-api-key"))


def check_ensemble_id_local_id(step):
    """Step: the ensemble ID and the local ensemble ID match"""
    eq_(step.bigml["local_ensemble"].resource_id, world.ensemble["resource"])


def i_export_logistic_regression(step, filename):
    """Step: I export the logistic regression"""
    world.api.export(world.logistic_regression.get('resource'),
                     filename=res_filename(filename))


def i_create_local_logistic_regression_from_file(step, export_file):
    """Step: I create a local logistic regressin from file <export_file>"""
    step.bigml["local_logistic"] = LogisticRegression( \
        res_filename(export_file),
        api=BigML("wrong-user", "wrong-api-key"))


def check_logistic_regression_id_local_id(step):
    """Step: the logistic ID and the local logistic ID match"""
    eq_(step.bigml["local_logistic"].resource_id, world.logistic_regression["resource"])


def i_export_deepnet(step, filename):
    """Step: I export the deepnet"""
    world.api.export(world.deepnet.get('resource'),
                     filename=res_filename(filename))


def i_create_local_deepnet_from_file(step, export_file):
    """Step: I create a local deepnet from file <export_file>"""
    step.bigml["local_deepnet"] = Deepnet(res_filename(export_file),
                                  api=BigML("wrong-user", "wrong-api-key"))


def i_export_fusion(step, filename):
    """Step: I export the fusion"""
    world.api.export(world.fusion.get('resource'),
                     filename=res_filename(filename))


def i_create_local_fusion_from_file(step, export_file):
    """Step: I create a local fusion from file <export_file>"""
    step.bigml["local_fusion"] = Fusion( \
        res_filename(export_file), api=BigML("wrong-user", "wrong-api-key"))


def check_fusion_id_local_id(step):
    """Step: the fusion ID and the local fusion ID match"""
    eq_(step.bigml["local_fusion"].resource_id, world.fusion["resource"])


def i_export_linear_regression(step, filename):
    """Step: I export the linear regression"""
    world.api.export(world.linear_regression.get('resource'),
                     filename=res_filename(filename))


def i_create_local_linear_regression_from_file(step, export_file):
    """Step: I create a local linear regression from file <export_file>"""
    step.bigml["local_linear_regression"] = LinearRegression( \
        res_filename(export_file), api=BigML("wrong-user", "wrong-api-key"))


def check_linear_regression_id_local_id(step):
    """Step: the linear regression ID and the local linear regression ID
    match
    """
    eq_(step.bigml["local_linear_regression"].resource_id,
        world.linear_regression["resource"])


def local_logistic_prediction_is(step, input_data, prediction):
    """Checking local logistic prediction"""
    eq_(step.bigml["local_logistic"].predict(input_data), prediction)


def local_linear_prediction_is(step, input_data, prediction):
    """Checking local linear prediction"""
    eq_(step.bigml["local_linear_regression"].predict(input_data),
        prediction, precision=5)

def local_deepnet_prediction_is(step, input_data, prediction):
    """Checking local deepnet prediction"""
    eq_(step.bigml["local_deepnet"].predict(input_data), prediction, precision=4)


def local_ensemble_prediction_is(step, input_data, prediction):
    """Checking local ensemble prediction"""
    eq_(step.bigml["local_ensemble"].predict(input_data), prediction, precision=5)


def local_model_prediction_is(step, input_data, prediction):
    """Checking local model prediction"""
    eq_(step.bigml["local_model"].predict(input_data), prediction, precision=5)


def local_cluster_prediction_is(step, input_data, prediction):
    """Checking local cluster prediction"""
    eq_(step.bigml["local_cluster"].centroid(input_data), prediction)


def local_anomaly_prediction_is(step, input_data, prediction):
    """Checking local anomaly prediction"""
    eq_(step.bigml["local_anomaly"].anomaly_score(input_data), prediction)


def local_association_prediction_is(step, input_data, prediction):
    """Checking local association prediction"""
    eq_(step.bigml["local_association"].association_set(input_data), prediction)


def local_time_series_prediction_is(step, input_data, prediction):
    """Checking local time series prediction"""
    eq_(step.bigml["local_time_series"].centroid(input_data), prediction)


def clone_model(step, model):
    """Step: I clone model
    """
    resource = world.api.clone_model(model, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.model = resource['object']
    # save reference
    world.models.append(resource['resource'])

def the_cloned_model_is(step, model):
    """Checking the model is a clone"""
    eq_(world.model["origin"], model)


def clone_deepnet(step, deepnet):
    """Step: I clone deepnet"""
    resource = world.api.clone_deepnet(deepnet, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.deepnet = resource['object']
    # save reference
    world.deepnets.append(resource['resource'])


def the_cloned_deepnet_is(step, deepnet):
    """Checking the deepnet is a clone"""
    eq_(world.deepnet["origin"], deepnet)


def clone_logistic_regression(step, logistic_regression):
    """Step: I clone logistic regression"""
    resource = world.api.clone_logistic_regression(
        logistic_regression, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.logistic_regression = resource['object']
    # save reference
    world.logistic_regressions.append(resource['resource'])


def the_cloned_logistic_regression_is(step, logistic_regression):
    """Checking logistic regression is a clone"""
    eq_(world.logistic_regression["origin"], logistic_regression)


def check_deepnet_id_local_id(step):
    """Checking that deepnet ID and local deepnet ID match"""
    eq_(world.deepnet["resource"], step.bigml["local_deepnet"].resource_id)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from bigml.api import HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_


def i_create_project(step, name):
    """Creating projects """
    resource = world.api.create_project({"name": name})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.project = resource['object']
    # save reference
    world.projects.append(resource['resource'])


def the_project_is_finished(step, secs):
    """Waiting for project to be finished"""
    wait_until_status_code_is(FINISHED, FAULTY, secs, world.project)


def i_update_project_name_with(step, name=""):
    """Updating project name"""
    resource = world.api.update_project(world.project.get('resource'),
                                        {"name": name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.project = resource['object']


def i_check_project_name(step, name=""):
    """Checking project name"""
    updated_name = world.project.get("name", "")
    eq_(updated_name, name)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012, 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from bigml.api import HTTP_CREATED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_, ok_

def i_create_an_evaluation(step, shared=None):
    """Step: I create an evaluation for the model with the dataset"""
    dataset = world.dataset.get('resource')
    model = world.model.get('resource')
    resource = world.api.create_evaluation(model, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


def i_create_an_evaluation_ensemble(step, params=None):
    """Step: I create an evaluation for the ensemble with the dataset"""
    if params is None:
        params = {}
    dataset = world.dataset.get('resource')
    ensemble = world.ensemble.get('resource')
    resource = world.api.create_evaluation(ensemble, dataset, params)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


def i_create_an_evaluation_logistic(step):
    """Step: I create an evaluation for the logistic regression with
    the dataset
    """
    dataset = world.dataset.get('resource')
    logistic = world.logistic_regression.get('resource')
    resource = world.api.create_evaluation(logistic, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


def i_create_an_evaluation_deepnet(step):
    """Step: I create an evaluation for the deepnet with the dataset"""
    dataset = world.dataset.get('resource')
    deepnet = world.deepnet.get('resource')
    resource = world.api.create_evaluation(deepnet, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


def i_create_an_evaluation_fusion(step):
    """Step: I create an evaluation for the fusion with the dataset"""
    dataset = world.dataset.get('resource')
    fusion = world.fusion.get('resource')
    resource = world.api.create_evaluation(fusion, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.evaluation = resource['object']
    world.evaluations.append(resource['resource'])


def wait_until_evaluation_status_code_is(step, code1, code2, secs):
    """Step: I wait until the evaluation status code is either <code1> or
    <code2> less than <secs>"""
    world.evaluation = wait_until_status_code_is(
        code1, code2, secs, world.evaluation)


def the_evaluation_is_finished_in_less_than(step, secs):
    """Step: I wait until the evaluation is ready less than <secs>"""
    wait_until_evaluation_status_code_is(step, FINISHED, FAULTY, secs)


def the_measured_measure_is_value(step, measure, value):
    """Step: the measured <measure> is <value>"""
    ev_ = world.evaluation['result']['model'][measure] + 0.0
    eq_(ev_, float(value), "The %s is: %s and %s is expected" % (
        measure, ev_, float(value)))


def the_measured_measure_is_greater_value(step, measure, value):
    """#@step(r'the measured <measure> is greater than <value>"""
    ok_(float(world.evaluation['result']['model'][measure]) > float(value))



# -*- coding: utf-8 -*-
# pylint: disable=super-init-not-called
#
# Copyright 2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A local Predictive model class abstracting all kind of models

This module abstracts any BigML model to make predictions locally or
embedded into your application without needing to send requests to
BigML.io.

This module cannot only save you a few credits, but also enormously
reduce the latency for each prediction and let you use your supervised models
offline.

Example usage (assuming that you have previously set up the BIGML_USERNAME
and BIGML_API_KEY environment variables and that you own the
logisticregression/id below):

from bigml.api import BigML
from bigml.local_model import LocalModel

api = BigML()

model = LocalModel(
    'logisticregression/5026965515526876630001b2')
model.predict({"petal length": 3, "petal width": 1,
               "sepal length": 1, "sepal width": 0.5})

"""

import json
import os


from bigml.api import get_resource_id, get_resource_type, \
    get_api_connection, get_ensemble_id
from bigml.basemodel import BaseModel
from bigml.model import Model
from bigml.ensemble import Ensemble
from bigml.logistic import LogisticRegression
from bigml.deepnet import Deepnet
from bigml.linear import LinearRegression
from bigml.fusion import Fusion
from bigml.cluster import Cluster
from bigml.anomaly import Anomaly
from bigml.association import Association
try:
    from bigml.topicmodel import TopicModel
    TOPIC_ENABLED = True
except ImportError:
    TOPIC_ENABLED = False
from bigml.pca import PCA
from bigml.constants import OUT_NEW_FIELDS, OUT_NEW_HEADERS, INTERNAL
from bigml.util import get_data_format, get_formatted_data, format_data


SUPERVISED_CLASSES = {
    "model": Model,
    "ensemble": Ensemble,
    "logisticregression": LogisticRegression,
    "deepnet": Deepnet,
    "linearregression": LinearRegression,
    "fusion": Fusion}


DFT_OUTPUTS = ["prediction", "probability"]


MODEL_CLASSES = {
    "cluster": Cluster,
    "anomaly": Anomaly,
    "association": Association,
    "topicmodel": TopicModel,
    "pca": PCA}
MODEL_CLASSES.update(SUPERVISED_CLASSES)


def extract_id(model, api):
    """Extract the resource id from:
        - a resource ID string
        - a list of resources (ensemble +  models)
        - a resource structure
        - the name of the file that contains a resource structure

    """
    # the string can be a path to a JSON file
    if isinstance(model, str):
        try:
            path = os.path.dirname(os.path.abspath(model))
            with open(model) as model_file:
                model = json.load(model_file)
                resource_id = get_resource_id(model)
                if resource_id is None:
                    raise ValueError("The JSON file does not seem"
                                     " to contain a valid BigML resource"
                                     " representation.")
                api.storage = path
        except IOError:
            # if it is not a path, it can be a model id
            resource_id = get_resource_id(model)
            if resource_id is None:
                for resource_type in MODEL_CLASSES.keys():
                    if model.find("%s/" % resource_type) > -1:
                        raise Exception(
                            api.error_message(model,
                                              resource_type=resource_type,
                                              method="get"))
                raise IOError("Failed to open the expected JSON file"
                              " at %s." % model)
        except ValueError:
            raise ValueError("Failed to interpret %s."
                             " JSON file expected.")
    if isinstance(model, list):
        resource_id = get_ensemble_id(model[0])
        if resource_id is None:
            raise ValueError("The first argument does not contain a valid"
                             " BigML model structure.")
    else:
        resource_id = get_resource_id(model)
        if resource_id is None:
            raise ValueError("The first argument does not contain a valid"
                             " BigML model structure.")
    return resource_id, model


class LocalModel(BaseModel):
    """ A lightweight wrapper around any BigML model.

    Uses any BigML remote supervised model to build a local version
    that can be used to generate predictions locally.

    """

    def __init__(self, model, api=None, cache_get=None,
                 operation_settings=None):

        self.api = get_api_connection(api)
        resource_id, model = extract_id(model, self.api)
        resource_type = get_resource_type(resource_id)
        if resource_type == "topicmodel" and not TOPIC_ENABLED:
            raise ValueError("Failed to import the TopicModel class. "
                "Please, check the bindings extra options to install"
                " the class.")
        kwargs = {"api": self.api, "cache_get": cache_get}
        if resource_type in SUPERVISED_CLASSES.keys() and \
                resource_type != "linearregression":
            kwargs.update({"operation_settings": operation_settings})
        local_model = MODEL_CLASSES[resource_type](model, **kwargs)
        self.__class__.__bases__ = local_model.__class__.__bases__
        for attr, value in list(local_model.__dict__.items()):
            setattr(self, attr, value)
        self.local_model = local_model
        self.supervised = resource_type in SUPERVISED_CLASSES.keys()
        self.name = self.local_model.name
        self.description = self.local_model.description

    def predict(self, *args, **kwargs):
        """Delegating method to local model object"""
        return self.local_model.predict(*args, **kwargs)

    def data_transformations(self):
        """Returns the pipeline transformations previous to the modeling
        step as a pipeline, so that they can be used in local predictions.
        """
        return self.local_model.data_transformations()

    def batch_predict(self, input_data_list, outputs=None, **kwargs):
        """Creates a batch prediction for a list of inputs using the local
        supervised model. Allows to define some output settings to
        decide the fields to be added to the input_data (prediction,
        probability, etc.) and the name that we want to assign to these new
        fields. The outputs argument accepts a dictionary with keys
        "output_fields", to contain a list of the prediction properties to add
        (["prediction", "probability"] by default) and "output_headers", to
        contain a list of the headers to be used when adding them (identical
        to "output_fields" list, by default).

        :param input_data_list: List of input data to be predicted
        :type input_data_list: list or Panda's dataframe
        :param dict outputs: properties that define the headers and fields to
                             be added to the input data
        :return: the list of input data plus the predicted values
        :rtype: list or Panda's dataframe depending on the input type in
                input_data_list
        """
        if isinstance(self.local_model, Association):
            raise("The method is not available for Associations.")
        if self.supervised:
            if outputs is None:
                outputs = {}
            new_fields = outputs.get(OUT_NEW_FIELDS, DFT_OUTPUTS)
            new_headers = outputs.get(OUT_NEW_HEADERS, new_fields)
            if len(new_fields) > len(new_headers):
                new_headers.expand(new_fields[len(new_headers):])
            else:
                new_headers = new_headers[0: len(new_fields)]
            data_format = get_data_format(input_data_list)
            inner_data_list = get_formatted_data(input_data_list, INTERNAL)
            kwargs.update({"full": True})
            for input_data in inner_data_list:
                prediction = self.predict(input_data, **kwargs)
                for index, key in enumerate(new_fields):
                    try:
                        input_data[new_headers[index]] = prediction[key]
                    except KeyError:
                        pass
            if data_format != INTERNAL:
                return format_data(inner_data_list, out_format=data_format)
            return inner_data_list
        else:
            return self.local_model.batch_predict(input_data_list,
                outputs=outputs, **kwargs)

    #pylint: disable=locally-disabled,arguments-differ
    def dump(self, **kwargs):
        """Delegate to local model"""
        self.local_model.dump(**kwargs)

    def dumps(self):
        """Delegate to local model"""
        return self.local_model.dumps()
